# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""DataLoader for Searcher task."""

from typing import AnyStr, List, Optional

import numpy as np
from tensorflow_examples.lite.model_maker.core.api.api_util import mm_export


@mm_export("searcher.DataLoader")
class DataLoader(object):
  """Base DataLoader class for Searcher task."""

  def __init__(
      self,
      embedder_path: Optional[str] = None,
      dataset: Optional[np.ndarray] = None,
      metadata: Optional[List[AnyStr]] = None,
  ) -> None:
    """Initializes DataLoader for Searcher task.

    Args:
      embedder_path: Path to the TFLite Embedder model file.
      dataset: Embedding dataset used to build on-device ScaNN index file. The
        dataset shape should be (dataset_size, embedding_dim). If None,
        `dataset` will be generated from raw input data later.
      metadata:  The metadata for each data in the dataset. The length of
        `metadata` should be same as `dataset` and passed in the same order as
        `dataset`. If `dataset` is set, `metadata` should be set as well.
    """
    self._embedder_path = embedder_path

    # Cache dataset list which can be concatenated to the single dataset. This
    # is used since users may load the data several times, if each time the data
    # are directly concatenated together, memory copy will be costed each time.
    self._cache_dataset_list = []

    if dataset is None:
      # Sets dataset and metadata as empty. Will load them from raw input data
      # later.
      self._dataset = np.array([])
      self._metadata = []
    else:
      # Directly sets dataset and metadata.
      self._dataset = dataset
      self._metadata = metadata

  def __len__(self):
    return len(self.dataset)

  @property
  def dataset(self) -> np.ndarray:
    """Gets the dataset.

    Due to performance consideration, we don't return a copy, but the returned
    `self._dataset` should never be changed.
    """
    if self._cache_dataset_list:
      # Concatenates the `self._dataset` and the datasets in
      # `self._cache_dataset_list`.
      if self._dataset.size > 0:
        dataset_list = [self._dataset] + self._cache_dataset_list
      else:
        dataset_list = self._cache_dataset_list

      self._dataset = np.vstack(dataset_list)
      self._cache_dataset_list = []
    return self._dataset

  @property
  def metadata(self) -> List[AnyStr]:
    """Gets the metadata."""
    return self._metadata

  @property
  def embedder_path(self) -> Optional[str]:
    """Gets the path to the TFLite Embedder model file."""
    return self._embedder_path

  def append(self, data_loader: "DataLoader") -> None:
    """Appends the dataset.

    Don't check if embedders from the two data loader are the same in this
    function. Users are responsible to keep the embedder identical.

    Args:
      data_loader: The data loader in which the data will be appended.
    """
    if self._dataset.shape[1] != data_loader.dataset.shape[1]:
      raise ValueError("The embedding dimension must be the same.")

    # Appends the array.
    self._cache_dataset_list.append(data_loader.dataset)
    self._metadata = self._metadata + data_loader.metadata



# Copyright 2021 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""A simple script to allow pretraining of the efficientnet backbone."""

import os
import re

from absl import app
from absl import flags
from absl import logging
import tensorflow as tf
import tensorflow_datasets as tfds

from tensorflow_examples.lite.model_maker.third_party.efficientdet.backbone import efficientnet_builder
from tensorflow_examples.lite.model_maker.third_party.efficientdet.backbone import efficientnet_model
from tensorflow_examples.lite.model_maker.third_party.efficientdet.backbone import preprocessing

flags.DEFINE_string(
    'tpu',
    default=None,
    help='The Cloud TPU to use for training. This should be either the name '
    'used when creating the Cloud TPU, or a grpc://ip.address.of.tpu:8470 '
    'url.')
flags.DEFINE_string(
    'gcp_project',
    default=None,
    help='Project name for the Cloud TPU-enabled project. If not specified, '
    'we will attempt to automatically detect the GCE project from metadata.')
flags.DEFINE_string(
    'tpu_zone',
    default=None,
    help='GCE zone where the Cloud TPU is located in. If not specified, we '
    'will attempt to automatically detect the GCE project from metadata.')

flags.DEFINE_enum('strategy', '', ['tpu', 'gpus', ''],
                  'Training: gpus for multi-gpu, if None, use TF default.')
flags.DEFINE_string(
    'model_dir', None, 'The directory where the model weights and '
    'training/evaluation summaries are stored.')
flags.DEFINE_string('model_name', 'efficientnet-b0', 'Model name.')
flags.DEFINE_integer('batch_size', 64, 'training batch size')
flags.DEFINE_integer('num_epochs', 10, 'Number of epochs to train.')
flags.DEFINE_float('learning_rate', 1e-4, 'The learning rate to use.')
flags.DEFINE_float(
    'label_smoothing', 0.1,
    'Label smoothing parameter used in the softmax_cross_entropy')
flags.DEFINE_float('weight_decay', 1e-5,
                   'Weight decay coefficiant for l2 regularization.')
flags.DEFINE_string(
    'augment_name', None, '`string` that is the name of the augmentation method'
    'to apply to the image. `autoaugment` if AutoAugment is to be used or'
    '`randaugment` if RandAugment is to be used. If the value is `None` no'
    'augmentation method will be applied applied. See autoaugment.py for  '
    'more details.')

flags.DEFINE_integer(
    'randaug_num_layers', 2,
    'If RandAug is used, what should the number of layers be.'
    'See autoaugment.py for detailed description.')

flags.DEFINE_integer(
    'randaug_magnitude', 10,
    'If RandAug is used, what should the magnitude be. '
    'See autoaugment.py for detailed description.')

FLAGS = flags.FLAGS


def create_dataset(dataset: tf.data.Dataset, num_classes: int,
                   is_training: bool) -> tf.data.Dataset:
  """Produces a full, augmented dataset from the inptu dataset."""
  _, _, resolution, _ = efficientnet_builder.efficientnet_params(
      FLAGS.model_name)

  def process_data(image, label):
    image = preprocessing.preprocess_image(
        image,
        is_training=is_training,
        use_bfloat16=FLAGS.strategy == 'tpus',
        image_size=resolution,
        augment_name=FLAGS.augment_name,
        randaug_num_layers=FLAGS.randaug_num_layers,
        randaug_magnitude=FLAGS.randaug_magnitude,
        resize_method=None)

    label = tf.one_hot(label, num_classes)
    return image, label

  dataset = dataset.map(
      process_data, num_parallel_calls=tf.data.experimental.AUTOTUNE)
  dataset = dataset.batch(FLAGS.batch_size, drop_remainder=True)
  dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)

  return dataset


class TrainableModel(efficientnet_model.Model):
  """Wraps efficientnet to make a keras trainable model.

  Handles efficientnet's multiple outputs and adds weight decay.
  """

  def __init__(self,
               blocks_args=None,
               global_params=None,
               name=None,
               weight_decay=0.0):
    super().__init__(
        blocks_args=blocks_args, global_params=global_params, name=name)

    self.weight_decay = weight_decay

  def _reg_l2_loss(self, weight_decay, regex=r'.*(kernel|weight):0$'):
    """Return regularization l2 loss loss."""
    var_match = re.compile(regex)
    return weight_decay * tf.add_n([
        tf.nn.l2_loss(v)
        for v in self.trainable_variables
        if var_match.match(v.name)
    ])

  def train_step(self, data):
    images, labels = data

    with tf.GradientTape() as tape:
      pred = self(images, training=True)[0]
      loss = self.compiled_loss(
          labels,
          pred,
          regularization_losses=[self._reg_l2_loss(self.weight_decay)])

    self.optimizer.minimize(loss, self.trainable_variables, tape=tape)
    self.compiled_metrics.update_state(labels, pred)
    return {m.name: m.result() for m in self.metrics}

  def test_step(self, data):
    images, labels = data
    pred = self(images, training=False)[0]

    self.compiled_loss(
        labels,
        pred,
        regularization_losses=[self._reg_l2_loss(self.weight_decay)])

    self.compiled_metrics.update_state(labels, pred)
    return {m.name: m.result() for m in self.metrics}


def main(_) -> None:
  if FLAGS.strategy == 'tpu':
    tpu_cluster_resolver = tf.distribute.cluster_resolver.TPUClusterResolver(
        FLAGS.tpu, zone=FLAGS.tpu_zone, project=FLAGS.gcp_project)
    tf.config.experimental_connect_to_cluster(tpu_cluster_resolver)
    tf.tpu.experimental.initialize_tpu_system(tpu_cluster_resolver)
    ds_strategy = tf.distribute.TPUStrategy(tpu_cluster_resolver)
    logging.info('All devices: %s', tf.config.list_logical_devices('TPU'))
  elif FLAGS.strategy == 'gpus':
    ds_strategy = tf.distribute.MirroredStrategy()
    logging.info('All devices: %s', tf.config.list_physical_devices('GPU'))
  else:
    if tf.config.list_physical_devices('GPU'):
      ds_strategy = tf.distribute.OneDeviceStrategy('device:GPU:0')
    else:
      ds_strategy = tf.distribute.OneDeviceStrategy('device:CPU:0')

  with ds_strategy.scope():
    (ds_train, ds_val), ds_info = tfds.load(
        'imagenet2012',
        split=['train', 'validation'],
        shuffle_files=True,
        as_supervised=True,
        with_info=True,
        decoders={
            'image': tfds.decode.SkipDecoding(),
        },
    )

    num_classes = ds_info.features['label'].num_classes

    ds_train = create_dataset(ds_train, num_classes, is_training=True)

    ds_val = create_dataset(ds_val, num_classes, is_training=False)

    blocks_args, global_params = efficientnet_builder.get_model_params(
        FLAGS.model_name,
        override_params={'num_classes': ds_info.features['label'].num_classes})
    model = TrainableModel(blocks_args, global_params, FLAGS.model_name,
                           FLAGS.weight_decay)

    steps_per_epoch = ds_info.splits['train'].num_examples // FLAGS.batch_size
    total_steps = steps_per_epoch * FLAGS.num_epochs
    cosign_decay = tf.keras.experimental.CosineDecay(FLAGS.learning_rate,
                                                     total_steps)
    model.compile(
        optimizer=tf.keras.optimizers.Adam(cosign_decay),
        loss=tf.keras.losses.CategoricalCrossentropy(
            label_smoothing=FLAGS.label_smoothing, from_logits=True),
        metrics=[tf.keras.metrics.CategoricalAccuracy()],
    )

    ckpt_callback = tf.keras.callbacks.ModelCheckpoint(
        os.path.join(FLAGS.model_dir, 'ckpt-{epoch:d}'),
        verbose=1,
        save_weights_only=True)
    tb_callback = tf.keras.callbacks.TensorBoard(
        log_dir=FLAGS.model_dir, update_freq=100)
    rstr_callback = tf.keras.callbacks.experimental.BackupAndRestore(
        backup_dir=FLAGS.model_dir)

    model.fit(
        ds_train,
        epochs=FLAGS.num_epochs,
        validation_data=ds_val,
        callbacks=[ckpt_callback, tb_callback, rstr_callback],
        # don't log spam if running on tpus
        verbose=2 if FLAGS.strategy == 'tpu' else 1,
    )


if __name__ == '__main__':
  app.run(main)



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Image DataLoader for Searcher task."""

import logging
import os

import numpy as np
import tensorflow as tf
from tensorflow_examples.lite.model_maker.core.api.api_util import mm_export
from tensorflow_examples.lite.model_maker.core.data_util import metadata_loader
from tensorflow_examples.lite.model_maker.core.data_util import searcher_dataloader
from tensorflow_lite_support.python.task.core import base_options as base_options_module
from tensorflow_lite_support.python.task.processor.proto import embedding_options_pb2
from tensorflow_lite_support.python.task.vision import image_embedder
from tensorflow_lite_support.python.task.vision.core import tensor_image

_MetadataType = metadata_loader.MetadataType
_BaseOptions = base_options_module.BaseOptions


@mm_export("searcher.ImageDataLoader")
class DataLoader(searcher_dataloader.DataLoader):
  """DataLoader class for Image Searcher Task."""

  def __init__(
      self,
      embedder: image_embedder.ImageEmbedder,
      metadata_type: _MetadataType = _MetadataType.FROM_FILE_NAME) -> None:
    """Initializes DataLoader for Image Searcher task.

    Args:
      embedder: Embedder to generate embedding from raw input image.
      metadata_type: Type of MetadataLoader to load metadata for each input
        data. By default, load the file name as metadata for each input data.
    """
    self._embedder = embedder
    super().__init__(embedder_path=embedder.options.base_options.file_name)

    # Creates the metadata loader.
    self.metadata_type = metadata_type
    if metadata_type is _MetadataType.FROM_FILE_NAME:
      self._metadata_loader = metadata_loader.MetadataLoader.from_file_name()
    elif metadata_type is _MetadataType.FROM_DAT_FILE:
      self._metadata_loader = metadata_loader.MetadataLoader.from_dat_file()
    else:
      raise ValueError("Unsuported metadata_type.")

  @classmethod
  def create(cls,
             image_embedder_path: str,
             metadata_type: _MetadataType = _MetadataType.FROM_FILE_NAME,
             l2_normalize: bool = False) -> "DataLoader":
    """Creates DataLoader for the Image Searcher task.

    Args:
      image_embedder_path: Path to the ".tflite" image embedder model.
      metadata_type: Type of MetadataLoader to load metadata for each input
        image based on image path. By default, load the file name as metadata
        for each input image.
      l2_normalize: Whether to normalize the returned feature vector with L2
        norm. Use this option only if the model does not already contain a
        native L2_NORMALIZATION TF Lite Op. In most cases, this is already the
        case and L2 norm is thus achieved through TF Lite inference.

    Returns:
      DataLoader object created for the Image Searcher task.
    """
    # Creates ImageEmbedder.
    image_embedder_path = os.path.abspath(image_embedder_path)
    with tf.io.gfile.GFile(image_embedder_path, "rb") as f:
      image_embedder_content = f.read()
    base_options = _BaseOptions(
        file_content=image_embedder_content, file_name=image_embedder_path)
    embedding_options = embedding_options_pb2.EmbeddingOptions(
        l2_normalize=l2_normalize)
    options = image_embedder.ImageEmbedderOptions(
        base_options=base_options, embedding_options=embedding_options)
    embedder = image_embedder.ImageEmbedder.create_from_options(options)

    return cls(embedder, metadata_type)

  def load_from_folder(self, path: str, mode: str = "r") -> None:
    """Loads image data from folder.

    Users can load images from different folders one by one. For instance,
    ```
    # Creates data_loader instance.
    data_loader = image_searcher_dataloader.DataLoader.create(tflite_path)

    # Loads images, first from `image_path1` and secondly from `image_path2`.
    data_loader.load_from_folder(image_path1)
    data_loader.load_from_folder(image_path2)
    ```

    Args:
      path: image directory to be loaded.
      mode: mode in which the file is opened, Used when metadata_type is
        FROM_DAT_FILE. Only 'r' and 'rb' are supported. 'r' means opening for
        reading, 'rb' means opening for reading binary.
    """
    embedding_list = []
    metadata_list = []

    i = 0
    # Gets the image files in the folder and loads images.
    for root, _, files in tf.io.gfile.walk(path):
      for name in files:
        image_path = os.path.join(root, name)
        if image_path.lower().endswith(".dat"):
          continue

        try:
          with tf.io.gfile.GFile(image_path, "rb") as f:
            buffer = f.read()
          image = tensor_image.TensorImage.create_from_buffer(buffer)
        except RuntimeError as e:
          logging.warning(
              "Can't read image from the image path %s with the error %s",
              image_path, e)
          continue

        try:
          embedding = self._embedder.embed(
              image).embeddings[0].feature_vector.value
        except (RuntimeError, ValueError) as e:
          logging.warning("Can't get the embedding of %s with the error %s",
                          image_path, e)
          continue

        embedding_list.append(embedding)
        if self.metadata_type == _MetadataType.FROM_DAT_FILE:
          metadata = self._metadata_loader.load(image_path, mode=mode)
        else:
          metadata = self._metadata_loader.load(image_path)
        metadata_list.append(metadata)

        i += 1
        if i % 1000 == 0:
          logging.info("Processed %d images.", i)

    cache_dataset = np.stack(embedding_list)
    self._cache_dataset_list.append(cache_dataset)
    self._metadata = self._metadata + metadata_list



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Recommendation dataloader class."""

import collections
import functools
import json
import os

import tensorflow as tf

from tensorflow_examples.lite.model_maker.core import file_util
from tensorflow_examples.lite.model_maker.core.api import mm_export
from tensorflow_examples.lite.model_maker.core.data_util import dataloader
from tensorflow_examples.lite.model_maker.core.data_util import recommendation_config
from tensorflow_examples.lite.model_maker.third_party.recommendation.ml.data import example_generation_movielens as _gen
from tensorflow_examples.lite.model_maker.third_party.recommendation.ml.model import input_pipeline
from tensorflow_examples.lite.model_maker.third_party.recommendation.ml.model import utils


@mm_export('recommendation.DataLoader')
class RecommendationDataLoader(dataloader.DataLoader):
  """Recommendation data loader."""

  def __init__(self, dataset, size, vocab):
    """Init data loader.

    Dataset is tf.data.Dataset of examples, containing:
      for inputs:
      - 'context': int64[], context ids as the input of variable length.
      for outputs:
      - 'label': int64[1], label id to predict.
    where context is controlled by `max_context_length` in generating examples.

    The vocab should be a dict maps `id` to `item`, where:
    - id: int
    - item: a vocab entry. For example, for movielens, the item is a dict:
        {'id': int, 'title': str, 'genres': list[str], 'count': int}

    Args:
      dataset: tf.data.Dataset for recommendation.
      size: int, dataset size.
      vocab: list of dict, each vocab item is described above.
    """
    super(RecommendationDataLoader, self).__init__(dataset, size)
    if not isinstance(vocab, dict):
      raise ValueError('Expect vocab to be a dict, but got: {}'.format(vocab))
    self.vocab = vocab
    self.max_vocab_id = max(self.vocab.keys())  # The max id in the vocab.

  def gen_dataset(self,
                  batch_size=1,
                  is_training=False,
                  shuffle=False,
                  input_pipeline_context=None,
                  preprocess=None,
                  drop_remainder=True,
                  total_steps=None):
    """Generates dataset, and overwrites default drop_remainder = True."""
    ds = super(RecommendationDataLoader, self).gen_dataset(
        batch_size=batch_size,
        is_training=is_training,
        shuffle=shuffle,
        input_pipeline_context=input_pipeline_context,
        preprocess=preprocess,
        drop_remainder=drop_remainder,
    )
    # TODO(tianlin): Consider to move the num_batches below to the super class.
    # Calculate steps by train data, if it is not set.
    if total_steps:
      num_batches = total_steps
    else:
      num_batches = self._size // batch_size
    return ds.take(num_batches)

  def split(self, fraction):
    return self._split(fraction, self.vocab)

  @classmethod
  def load_vocab(cls, vocab_file) -> collections.OrderedDict:
    """Loads vocab from file.

    The vocab file should be json format of: a list of list[size=4], where the 4
    elements are ordered as:
      [id=int, title=str, genres=str joined with '|', count=int]
    It is generated when preparing movielens dataset.

    Args:
      vocab_file: str, path to vocab file.

    Returns:
      vocab: an OrderedDict maps id to item. Each item represents a movie
         {
           'id': int,
           'title': str,
           'genres': list[str],
           'count': int,
         }
    """
    with tf.io.gfile.GFile(vocab_file) as f:
      vocab_json = json.load(f)
      vocab = collections.OrderedDict()
      for v in vocab_json:
        item = {
            'id': int(v[0]),
            'title': v[1],
            'genres': v[2].split('|'),
            'count': int(v[3]),
        }
        vocab[item['id']] = item
      return vocab

  @classmethod
  def _read_dataset(cls, data_filepattern: str,
                    input_spec: recommendation_config.InputSpec,
                    vocab_file_dir: str) -> tf.data.Dataset:
    features_and_vocabs_by_name = input_pipeline.get_features_and_vocabs_by_name(
        input_spec, vocab_file_dir)
    if not input_spec.HasField('label_feature'):
      raise ValueError('Field label_feature is required.')
    input_files = utils.GetShardFilenames(data_filepattern)
    d = tf.data.TFRecordDataset(input_files)
    d = d.shuffle(len(input_files))
    decode_fn = functools.partial(
        input_pipeline.decode_example,
        features_and_vocabs_by_name=features_and_vocabs_by_name,
        label_feature_name=input_spec.label_feature.feature_name)
    return d.map(decode_fn, num_parallel_calls=tf.data.AUTOTUNE)

  @classmethod
  def download_and_extract_movielens(cls, download_dir):
    """Downloads and extracts movielens dataset, then returns extracted dir."""
    return _gen.download_and_extract_data(download_dir)

  @classmethod
  def generate_movielens_dataset(
      cls,
      data_dir,
      generated_examples_dir=None,
      train_filename='train_movielens_1m.tfrecord',
      test_filename='test_movielens_1m.tfrecord',
      vocab_filename='movie_vocab.json',
      meta_filename='meta.json',
      min_timeline_length=3,
      max_context_length=10,
      max_context_movie_genre_length=10,
      min_rating=None,
      train_data_fraction=0.9,
      build_vocabs=True,
  ):
    """Generate movielens dataset, and returns a dict contains meta.

    Args:
      data_dir: str, path to dataset containing (unzipped) text data.
      generated_examples_dir: str, path to generate preprocessed examples.
        (default: same as data_dir)
      train_filename: str, generated file name for training data.
      test_filename: str, generated file name for test data.
      vocab_filename: str, generated file name for vocab data.
      meta_filename: str, generated file name for meta data.
      min_timeline_length: int, min timeline length to split train/eval set.
      max_context_length: int, max context length as one input.
      max_context_movie_genre_length: int, max context length of movie genre as
        one input.
      min_rating: int or None, include examples with min rating.
      train_data_fraction: float, percentage of training data [0.0, 1.0].
      build_vocabs: boolean, whether to build vocabs.

    Returns:
      Dict, metadata for the movielens dataset. Containing keys:
        `train_file`, `train_size`, `test_file`, `test_size`, vocab_file`,
        `vocab_size`, etc.
    """
    if not generated_examples_dir:
      # By default, set generated examples dir to data_dir
      generated_examples_dir = data_dir
    train_file = os.path.join(generated_examples_dir, train_filename)
    test_file = os.path.join(generated_examples_dir, test_filename)
    meta_file = os.path.join(generated_examples_dir, meta_filename)
    # Create dataset and meta, only if they are not existed.
    if not all([os.path.exists(f) for f in (train_file, test_file, meta_file)]):
      stats = _gen.generate_datasets(
          data_dir,
          output_dir=generated_examples_dir,
          min_timeline_length=min_timeline_length,
          max_context_length=max_context_length,
          max_context_movie_genre_length=max_context_movie_genre_length,
          min_rating=min_rating,
          build_vocabs=build_vocabs,
          train_data_fraction=train_data_fraction,
          train_filename=train_filename,
          test_filename=test_filename,
          vocab_filename=vocab_filename,
      )
      file_util.write_json_file(meta_file, stats)
    meta = file_util.load_json_file(meta_file)
    return meta

  @classmethod
  def get_num_classes(cls, meta) -> int:
    """Gets number of classes.

    0 is reserved. Number of classes is Max Id + 1, e.g., if Max Id = 100,
    then classes are [0, 100], that is 101 classes in total.

    Args:
      meta: dict, containing meta['vocab_max_id'].

    Returns:
      Number of classes.
    """
    return meta['vocab_max_id'] + 1

  @classmethod
  def from_movielens(cls,
                     data_dir,
                     data_tag,
                     input_spec: recommendation_config.InputSpec,
                     generated_examples_dir=None,
                     min_timeline_length=3,
                     max_context_length=10,
                     max_context_movie_genre_length=10,
                     min_rating=None,
                     train_data_fraction=0.9,
                     build_vocabs=True,
                     train_filename='train_movielens_1m.tfrecord',
                     test_filename='test_movielens_1m.tfrecord',
                     vocab_filename='movie_vocab.json',
                     meta_filename='meta.json'):
    """Generates data loader from movielens dataset.

    The method downloads and prepares dataset, then generates for train/eval.

    For `movielens` data format, see:
    - function `_generate_fake_data` in `recommendation_testutil.py`
    - Or, zip file: http://files.grouplens.org/datasets/movielens/ml-1m.zip

    Args:
      data_dir: str, path to dataset containing (unzipped) text data.
      data_tag: str, specify dataset in {'train', 'test'}.
      input_spec: InputSpec, specify data format for input and embedding.
      generated_examples_dir: str, path to generate preprocessed examples.
        (default: same as data_dir)
      min_timeline_length: int, min timeline length to split train/eval set.
      max_context_length: int, max context length as one input.
      max_context_movie_genre_length: int, max context length of movie genre as
        one input.
      min_rating: int or None, include examples with min rating.
      train_data_fraction: float, percentage of training data [0.0, 1.0].
      build_vocabs: boolean, whether to build vocabs.
      train_filename: str, generated file name for training data.
      test_filename: str, generated file name for test data.
      vocab_filename: str, generated file name for vocab data.
      meta_filename: str, generated file name for meta data.

    Returns:
      Data Loader.
    """
    if data_tag not in ('train', 'test'):
      raise ValueError(
          'Expected data_tag is train or test, but got {}'.format(data_tag))
    if not generated_examples_dir:
      # By default, set generated examples dir to data_dir
      generated_examples_dir = data_dir
    meta = cls.generate_movielens_dataset(
        data_dir,
        generated_examples_dir,
        train_filename=train_filename,
        test_filename=test_filename,
        vocab_filename=vocab_filename,
        meta_filename=meta_filename,
        min_timeline_length=min_timeline_length,
        max_context_length=max_context_length,
        max_context_movie_genre_length=max_context_movie_genre_length,
        min_rating=min_rating,
        train_data_fraction=train_data_fraction,
        build_vocabs=build_vocabs,
    )
    vocab = cls.load_vocab(meta['vocab_file'])
    if data_tag == 'train':
      ds = cls._read_dataset(meta['train_file'], input_spec,
                             generated_examples_dir)
      return cls(ds, meta['train_size'], vocab)
    elif data_tag == 'test':
      ds = cls._read_dataset(meta['test_file'], input_spec,
                             generated_examples_dir)
      return cls(ds, meta['test_size'], vocab)



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""IoU utils for box regression with iou losses.

Distance-IoU Loss: Faster and Better Learning for Bounding Box Regression.
https://arxiv.org/pdf/1911.08287.pdf
"""
import math
from typing import Union, Text
import numpy as np
import tensorflow as tf
FloatType = Union[tf.Tensor, float, np.float32, np.float64]


def _get_v(b1_height: FloatType, b1_width: FloatType, b2_height: FloatType,
           b2_width: FloatType) -> tf.Tensor:
  """Get the consistency measurement of aspect ratio for ciou."""

  @tf.custom_gradient
  def _get_grad_v(height, width):
    """backpropogate gradient."""
    arctan = tf.atan(tf.math.divide_no_nan(b1_width, b1_height)) - tf.atan(
        tf.math.divide_no_nan(width, height))
    v = 4 * ((arctan / math.pi)**2)

    def _grad_v(dv):
      """Grad for eager mode."""
      gdw = dv * 8 * arctan * height / (math.pi**2)
      gdh = -dv * 8 * arctan * width / (math.pi**2)
      return [gdh, gdw]

    def _grad_v_graph(dv, variables):
      """Grad for graph mode."""
      gdw = dv * 8 * arctan * height / (math.pi**2)
      gdh = -dv * 8 * arctan * width / (math.pi**2)
      return [gdh, gdw], tf.gradients(v, variables, grad_ys=dv)

    if tf.compat.v1.executing_eagerly_outside_functions():
      return v, _grad_v
    return v, _grad_v_graph

  return _get_grad_v(b2_height, b2_width)


def _iou_per_anchor(pred_boxes: FloatType,
                    target_boxes: FloatType,
                    iou_type: Text = 'iou') -> tf.Tensor:
  """Computing the IoU for a single anchor.

  Args:
    pred_boxes: predicted boxes, with coordinate [y_min, x_min, y_max, x_max].
    target_boxes: target boxes, with coordinate [y_min, x_min, y_max, x_max].
    iou_type: one of ['iou', 'ciou', 'diou', 'giou'].

  Returns:
    IoU loss float `Tensor`.
  """
  # t_ denotes target boxes and p_ denotes predicted boxes.
  t_ymin, t_xmin, t_ymax, t_xmax = target_boxes
  p_ymin, p_xmin, p_ymax, p_xmax = pred_boxes

  zero = tf.convert_to_tensor(0.0, t_ymin.dtype)
  p_width = tf.maximum(zero, p_xmax - p_xmin)
  p_height = tf.maximum(zero, p_ymax - p_ymin)
  t_width = tf.maximum(zero, t_xmax - t_xmin)
  t_height = tf.maximum(zero, t_ymax - t_ymin)
  p_area = p_width * p_height
  t_area = t_width * t_height

  intersect_ymin = tf.maximum(p_ymin, t_ymin)
  intersect_xmin = tf.maximum(p_xmin, t_xmin)
  intersect_ymax = tf.minimum(p_ymax, t_ymax)
  intersect_xmax = tf.minimum(p_xmax, t_xmax)
  intersect_width = tf.maximum(zero, intersect_xmax - intersect_xmin)
  intersect_height = tf.maximum(zero, intersect_ymax - intersect_ymin)
  intersect_area = intersect_width * intersect_height

  union_area = p_area + t_area - intersect_area
  iou_v = tf.math.divide_no_nan(intersect_area, union_area)
  if iou_type == 'iou':
    return iou_v  # iou is the simplest form.

  enclose_ymin = tf.minimum(p_ymin, t_ymin)
  enclose_xmin = tf.minimum(p_xmin, t_xmin)
  enclose_ymax = tf.maximum(p_ymax, t_ymax)
  enclose_xmax = tf.maximum(p_xmax, t_xmax)

  assert iou_type in ('giou', 'diou', 'ciou')
  if iou_type == 'giou':  # giou is the generalized iou.
    enclose_width = tf.maximum(zero, enclose_xmax - enclose_xmin)
    enclose_height = tf.maximum(zero, enclose_ymax - enclose_ymin)
    enclose_area = enclose_width * enclose_height
    giou_v = iou_v - tf.math.divide_no_nan(
        (enclose_area - union_area), enclose_area)
    return giou_v

  assert iou_type in ('diou', 'ciou')
  p_center = tf.stack([(p_ymin + p_ymax) / 2, (p_xmin + p_xmax) / 2], axis=-1)
  t_center = tf.stack([(t_ymin + t_ymax) / 2, (t_xmin + t_xmax) / 2], axis=-1)
  euclidean = tf.linalg.norm(t_center - p_center, axis=-1)
  diag_length = tf.linalg.norm(
      tf.stack([enclose_ymax - enclose_ymin, enclose_xmax - enclose_xmin],
               axis=-1),
      axis=-1)
  diou_v = iou_v - tf.math.divide_no_nan(euclidean**2, diag_length**2)
  if iou_type == 'diou':  # diou is the distance iou.
    return diou_v

  assert iou_type == 'ciou'
  v = _get_v(p_height, p_width, t_height, t_width)
  alpha = tf.math.divide_no_nan(v, ((1 - iou_v) + v))
  return diou_v - alpha * v  # the last one is ciou.


def iou_loss(pred_boxes: FloatType,
             target_boxes: FloatType,
             iou_type: Text = 'iou') -> tf.Tensor:
  """A unified interface for computing various IoU losses.

  Let B and B_gt denotes the pred_box and B_gt is the target box (ground truth):

    IoU = |B & B_gt| / |B | B_gt|

    GIoU = IoU - |C - B U B_gt| / C, where C is the smallest box covering B and
    B_gt.

    DIoU = IoU - E(B, B_gt)^2 / c^2, E is the Euclidean distance of the center
    points of B and B_gt, and c is the diagonal length of the smallest box
    covering the two boxes

    CIoU = IoU - DIoU - a * v, where a is a positive trade-off parameter, and
    v measures the consistency of aspect ratio:
      v = (arctan(w_gt / h_gt) - arctan(w / h)) * 4 / pi^2
    where (w_gt, h_gt) and (w, h) are the width and height of the target and
    predicted box respectively.

  The returned loss is computed as 1 - one of {IoU, GIoU, DIoU, CIoU}.

  Args:
    pred_boxes: predicted boxes, with coordinate [y_min, x_min, y_max, x_max]*.
      It can be multiple anchors, with each anchor box has four coordinates.
    target_boxes: target boxes, with coordinate [y_min, x_min, y_max, x_max]*.
      It can be multiple anchors, with each anchor box has four coordinates.
    iou_type: one of ['iou', 'ciou', 'diou', 'giou'].

  Returns:
    IoU loss float `Tensor`.
  """
  if iou_type not in ('iou', 'ciou', 'diou', 'giou'):
    raise ValueError(
        'Unknown loss_type {}, not iou/ciou/diou/giou'.format(iou_type))

  pred_boxes = tf.convert_to_tensor(pred_boxes)
  target_boxes = tf.cast(target_boxes, pred_boxes.dtype)

  # t_ denotes target boxes and p_ denotes predicted boxes: (y, x, y_max, x_max)
  pred_boxes_list = tf.unstack(pred_boxes, None, axis=-1)
  target_boxes_list = tf.unstack(target_boxes, None, axis=-1)
  assert len(pred_boxes_list) == len(target_boxes_list)
  assert len(pred_boxes_list) % 4 == 0

  iou_loss_list = []
  for i in range(0, len(pred_boxes_list), 4):
    pred_boxes = pred_boxes_list[i:i + 4]
    target_boxes = target_boxes_list[i:i + 4]

    # Compute mask.
    t_ymin, t_xmin, t_ymax, t_xmax = target_boxes
    mask = tf.math.logical_and(t_ymax > t_ymin, t_xmax > t_xmin)
    mask = tf.cast(mask, t_ymin.dtype)
    # Loss should be mask * (1 - iou) = mask - masked_iou.
    pred_boxes = [b * mask for b in pred_boxes]
    target_boxes = [b * mask for b in target_boxes]
    iou_loss_list.append(
        mask *
        (1 - tf.squeeze(_iou_per_anchor(pred_boxes, target_boxes, iou_type))))  # pytype: disable=wrong-arg-types  # numpy-scalars
  if len(iou_loss_list) == 1:
    return iou_loss_list[0]
  return tf.reduce_sum(tf.stack(iou_loss_list), 0)



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Custom classification model that is already retained by data."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow.compat.v2 as tf
from tensorflow_examples.lite.model_maker.core import compat


def generate_elements(ds):
  """Generates elements from `tf.data.dataset`."""
  if compat.get_tf_behavior() == 2:
    for element in ds.as_numpy_iterator():
      yield element
  else:
    iterator = tf.compat.v1.data.make_one_shot_iterator(ds)
    next_element = iterator.get_next()
    with tf.compat.v1.Session() as sess:
      try:
        while True:
          yield sess.run(next_element)
      except tf.errors.OutOfRangeError:
        return



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Region Similarity Calculators for BoxLists.

Region Similarity Calculators compare a pairwise measure of similarity
between the boxes in two BoxLists.
"""
from abc import ABCMeta
from abc import abstractmethod

import tensorflow.compat.v1 as tf


def area(boxlist, scope=None):
  """Computes area of boxes.

  Args:
    boxlist: BoxList holding N boxes
    scope: name scope.

  Returns:
    a tensor with shape [N] representing box areas.
  """
  with tf.name_scope(scope, 'Area'):
    y_min, x_min, y_max, x_max = tf.split(
        value=boxlist.get(), num_or_size_splits=4, axis=1)
    return tf.squeeze((y_max - y_min) * (x_max - x_min), [1])


def intersection(boxlist1, boxlist2, scope=None):
  """Compute pairwise intersection areas between boxes.

  Args:
    boxlist1: BoxList holding N boxes
    boxlist2: BoxList holding M boxes
    scope: name scope.

  Returns:
    a tensor with shape [N, M] representing pairwise intersections
  """
  with tf.name_scope(scope, 'Intersection'):
    y_min1, x_min1, y_max1, x_max1 = tf.split(
        value=boxlist1.get(), num_or_size_splits=4, axis=1)
    y_min2, x_min2, y_max2, x_max2 = tf.split(
        value=boxlist2.get(), num_or_size_splits=4, axis=1)
    all_pairs_min_ymax = tf.minimum(y_max1, tf.transpose(y_max2))
    all_pairs_max_ymin = tf.maximum(y_min1, tf.transpose(y_min2))
    intersect_heights = tf.maximum(0.0, all_pairs_min_ymax - all_pairs_max_ymin)
    all_pairs_min_xmax = tf.minimum(x_max1, tf.transpose(x_max2))
    all_pairs_max_xmin = tf.maximum(x_min1, tf.transpose(x_min2))
    intersect_widths = tf.maximum(0.0, all_pairs_min_xmax - all_pairs_max_xmin)
    return intersect_heights * intersect_widths


def iou(boxlist1, boxlist2, scope=None):
  """Computes pairwise intersection-over-union between box collections.

  Args:
    boxlist1: BoxList holding N boxes
    boxlist2: BoxList holding M boxes
    scope: name scope.

  Returns:
    a tensor with shape [N, M] representing pairwise iou scores.
  """
  with tf.name_scope(scope, 'IOU'):
    intersections = intersection(boxlist1, boxlist2)
    areas1 = area(boxlist1)
    areas2 = area(boxlist2)
    unions = (
        tf.expand_dims(areas1, 1) + tf.expand_dims(areas2, 0) - intersections)
    return tf.where(
        tf.equal(intersections, 0.0),
        tf.zeros_like(intersections), tf.truediv(intersections, unions))


class RegionSimilarityCalculator(object):
  """Abstract base class for region similarity calculator."""
  __metaclass__ = ABCMeta

  def compare(self, boxlist1, boxlist2, scope=None):
    """Computes matrix of pairwise similarity between BoxLists.

    This op (to be overridden) computes a measure of pairwise similarity between
    the boxes in the given BoxLists. Higher values indicate more similarity.

    Note that this method simply measures similarity and does not explicitly
    perform a matching.

    Args:
      boxlist1: BoxList holding N boxes.
      boxlist2: BoxList holding M boxes.
      scope: Op scope name. Defaults to 'Compare' if None.

    Returns:
      a (float32) tensor of shape [N, M] with pairwise similarity score.
    """
    with tf.name_scope(scope, 'Compare', [boxlist1, boxlist2]) as scope:
      return self._compare(boxlist1, boxlist2)

  @abstractmethod
  def _compare(self, boxlist1, boxlist2):
    pass


class IouSimilarity(RegionSimilarityCalculator):
  """Class to compute similarity based on Intersection over Union (IOU) metric.

  This class computes pairwise similarity between two BoxLists based on IOU.
  """

  def _compare(self, boxlist1, boxlist2):
    """Compute pairwise IOU similarity between the two BoxLists.

    Args:
      boxlist1: BoxList holding N boxes.
      boxlist2: BoxList holding M boxes.

    Returns:
      A tensor with shape [N, M] representing pairwise iou scores.
    """
    return iou(boxlist1, boxlist2)
